//+------------------------------------------------------------------+
//|                                                  TestStatNCT.mq5 |
//|                        Copyright 2016, MetaQuotes Software Corp. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2016, MetaQuotes Software Corp."
#property link      "https://www.mql5.com"
#property version   "1.00"

#include <Math\Stat\NoncentralT.mqh>

//--- precision
double calc_precision_pdf=10E-15;
double calc_precision_cdf=10E-15;
double calc_precision_quantile=10E-14;

//--- test values (x,pdf,cdf) for Noncentral T distribution
const double noncentral_t_x_values[]={0.0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1.0};
const double noncentral_t_pdf_values[]=
  {
   4.92773299108631898092114435293e-15, //N[PDF[Noncentral T Distribution[10,8],0.0],30]
   1.13022616309487993266124014688e-14, //N[PDF[Noncentral T Distribution[10,8],0.1],30]
   2.64161256619123935657695034087e-14, //N[PDF[Noncentral T Distribution[10,8],0.2],30]
   6.28106243570814021661611528450e-14, //N[PDF[Noncentral T Distribution[10,8],0.3],30]
   1.51636983962647220239107618892e-13, //N[PDF[Noncentral T Distribution[10,8],0.4],30]
   3.70864175555732846819012024992e-13, //N[PDF[Noncentral T Distribution[10,8],0.5],30]
   9.16615229573759829315327016195e-13, //N[PDF[Noncentral T Distribution[10,8],0.6],30]
   2.28327725393016095515004668783e-12, //N[PDF[Noncentral T Distribution[10,8],0.7],30]
   5.71609303970754101789636971640e-12, //N[PDF[Noncentral T Distribution[10,8],0.8],30]
   1.43395037240077895541041912481e-11, //N[PDF[Noncentral T Distribution[10,8],0.9],30]
   3.59391445912257844885457911125e-11  //N[PDF[Noncentral T Distribution[10,8],1.0],30]
  };
const double noncentral_t_cdf_values[]=
  {
   6.22096057427178412351599517259e-16, //N[CDF[Noncentral T Distribution[10,8],0.0],30]
   1.38876e-15,                         //N[CDF[Noncentral T Distribution[10,8],0.1],30]
   3.16637e-15,                         //N[CDF[Noncentral T Distribution[10,8],0.2],30]
   7.36264e-15,                         //N[CDF[Noncentral T Distribution[10,8],0.3],30]
   1.74292e-14,                         //N[CDF[Noncentral T Distribution[10,8],0.4],30]
   4.19181e-14,                         //N[CDF[Noncentral T Distribution[10,8],0.5],30]
   1.02187e-13,                         //N[CDF[Noncentral T Distribution[10,8],0.6],30]
   2.5185e-13,                          //N[CDF[Noncentral T Distribution[10,8],0.7],30]
   6.2582e-13,                          //N[CDF[Noncentral T Distribution[10,8],0.8],30]
   1.56338e-12,                         //N[CDF[Noncentral T Distribution[10,8],0.9],30]
   3.91469e-12                          //N[CDF[Noncentral T Distribution[10,8],1.0],30]
  };
//+------------------------------------------------------------------+
//| GetCorrectDigits                                                 |
//+------------------------------------------------------------------+
int GetCorrectDigits(const double delta)
  {
   double d=MathAbs(delta);
//--- check if delta to small
   if(d<10E-30)
      return(30);
//--- check if delta is large
   if(d>=1.0)
      return(0);
   int correct_digits=0;
   while(MathAbs(d)<1.0)
     {
      d=d*10;
      correct_digits++;
     }
//---
   return(correct_digits-1);
  }
//+------------------------------------------------------------------+
//| Structure for precision information                              |
//+------------------------------------------------------------------+
struct precision_info
  {
   //--- precision info
   double            pdf_max_delta;
   double            cdf_max_delta;
   double            quantile_max_delta;
  };
//+------------------------------------------------------------------+
//| Test function for Noncentral T distribution                      |
//+------------------------------------------------------------------+
bool TestNoncentralTDistribution(string test_name,const double nu,const double delta,const double &x_values[],const double &pdf_values[],const double &cdf_values[],precision_info &precision)
  {
   PrintFormat("%s started",test_name);
//---
   int N=ArraySize(x_values);
   int error_code=0;
//--- arrays for calculated values
   double noncentral_t_pdf[];
   double noncentral_t_cdf[];
   double noncentral_t_quantile[];
//--- arrays for differences
   double delta_noncentral_t_pdf[];
   double delta_noncentral_t_cdf[];
   double delta_noncentral_t_quantile[];
//--- prepare arrays
   ArrayResize(noncentral_t_pdf,N);
   ArrayResize(noncentral_t_cdf,N);
   ArrayResize(noncentral_t_quantile,N);
   ArrayResize(delta_noncentral_t_pdf,N);
   ArrayResize(delta_noncentral_t_cdf,N);
   ArrayResize(delta_noncentral_t_quantile,N);

//--- test 1: single values
   PrintFormat("%s: Test 1: Calculations for single values",test_name);
   for(int i=0;i<N;i++)
     {
      //--- calculate values
      noncentral_t_pdf[i]=MathProbabilityDensityNoncentralT(x_values[i],nu,delta,error_code);
      noncentral_t_cdf[i]=MathCumulativeDistributionNoncentralT(x_values[i],nu,delta,error_code);
      noncentral_t_quantile[i]=MathQuantileNoncentralT(noncentral_t_cdf[i],nu,delta,error_code);
      //--- calculate difference
      delta_noncentral_t_pdf[i]=pdf_values[i]-noncentral_t_pdf[i];
      delta_noncentral_t_cdf[i]=cdf_values[i]-noncentral_t_cdf[i];
      delta_noncentral_t_quantile[i]=x_values[i]-noncentral_t_quantile[i];
      PrintFormat("1 %d, PDF=%.20e, CDF=%.20e, Q=%.20e,",i,noncentral_t_pdf[i],noncentral_t_cdf[i],noncentral_t_quantile[i]);
      //PrintFormat("1 %d, dPDF=%.20e, dCDF=%.20e, dQ=%.20e,",i,delta_noncentral_t_pdf[i],delta_noncentral_t_cdf[i],delta_noncentral_t_quantile[i]);
     }
//--- check calculated values
   for(int i=0;i<N;i++)
     {
      if(MathAbs(delta_noncentral_t_pdf[i])>calc_precision_pdf)
         return false;
      if(MathAbs(delta_noncentral_t_cdf[i])>calc_precision_cdf)
         return false;
      if(MathAbs(delta_noncentral_t_quantile[i])>calc_precision_quantile)
         return false;
     }

//--- test 2: calculations on arrays
   PrintFormat("%s: Test 2: Calculations on arrays",test_name);
//--- calculate pdf, cdf and quantiles
   if(!MathProbabilityDensityNoncentralT(x_values,nu,delta,noncentral_t_pdf))
      return false;
   if(!MathCumulativeDistributionNoncentralT(x_values,nu,delta,noncentral_t_cdf))
      return false;
   if(!MathQuantileNoncentralT(noncentral_t_cdf,nu,delta,noncentral_t_quantile))
      return false;
//--- calculate differences
   ArrayResize(delta_noncentral_t_pdf,N);
   ArrayResize(delta_noncentral_t_cdf,N);
   ArrayResize(delta_noncentral_t_quantile,N);
   for(int i=0;i<N;i++)
     {
      delta_noncentral_t_pdf[i]=pdf_values[i]-noncentral_t_pdf[i];
      delta_noncentral_t_cdf[i]=cdf_values[i]-noncentral_t_cdf[i];
      delta_noncentral_t_quantile[i]=x_values[i]-noncentral_t_quantile[i];
      PrintFormat("2 %d, PDF=%.20e, CDF=%.20e, Q=%.20e,",i,noncentral_t_pdf[i],noncentral_t_cdf[i],noncentral_t_quantile[i]);
      //PrintFormat("2 %d, dPDF=%.20e, dCDF=%.20e, dQ=%.20e,",i,delta_noncentral_t_pdf[i],delta_noncentral_t_cdf[i],delta_noncentral_t_quantile[i]);
     }
//--- calculate max error
   precision.pdf_max_delta=MathAbs(delta_noncentral_t_pdf[0]);
   precision.cdf_max_delta=MathAbs(delta_noncentral_t_cdf[0]);
   precision.quantile_max_delta=MathAbs(delta_noncentral_t_quantile[0]);
   for(int i=0;i<N;i++)
     {
      precision.pdf_max_delta=MathMax(precision.pdf_max_delta,MathAbs(delta_noncentral_t_pdf[i]));
      precision.cdf_max_delta=MathMax(precision.cdf_max_delta,MathAbs(delta_noncentral_t_cdf[i]));
      precision.quantile_max_delta=MathMax(precision.quantile_max_delta,MathAbs(delta_noncentral_t_quantile[i]));
     }
//--- check calculations
   for(int i=0;i<N;i++)
     {
      if(MathAbs(delta_noncentral_t_pdf[i])>calc_precision_pdf)
         return false;
      if(MathAbs(delta_noncentral_t_cdf[i])>calc_precision_cdf)
         return false;
      if(MathAbs(delta_noncentral_t_quantile[i])>calc_precision_quantile)
         return false;
     }

//--- test 3: calculations on arrays
   PrintFormat("%s: Test 3: R-like overloaded functions",test_name);
//--- parameters for R-like overloaded functions
   bool tail=true;
   bool log_mode=true;
//--- calculate pdf, cdf and quantiles
   if(!MathProbabilityDensityNoncentralT(x_values,nu,delta,log_mode,noncentral_t_pdf))
      return false;
   if(!MathCumulativeDistributionNoncentralT(x_values,nu,delta,tail,log_mode,noncentral_t_cdf))
      return false;
   if(!MathQuantileNoncentralT(noncentral_t_cdf,nu,delta,tail,log_mode,noncentral_t_quantile))
      return false;
//--- calculate differences
   if(tail==true)
     {
      //--- lower tail mode (cdf)
      if(log_mode==true)
        {
         //--- logarithmic values
         for(int i=0;i<N;i++)
           {
            delta_noncentral_t_pdf[i]=pdf_values[i]-MathExp(noncentral_t_pdf[i]);
            delta_noncentral_t_cdf[i]=cdf_values[i]-MathExp(noncentral_t_cdf[i]);
            delta_noncentral_t_quantile[i]=x_values[i]-noncentral_t_quantile[i];
           }
        }
      else
      for(int i=0;i<N;i++)
        {
         delta_noncentral_t_pdf[i]=pdf_values[i]-noncentral_t_pdf[i];
         delta_noncentral_t_cdf[i]=cdf_values[i]-noncentral_t_cdf[i];
         delta_noncentral_t_quantile[i]=x_values[i]-noncentral_t_quantile[i];
        }
     }
   else
     {
      //--- upper tail mode (1.0-cdf)
      if(log_mode==true)
        {
         //--- logarithmic values
         for(int i=0;i<N;i++)
           {
            delta_noncentral_t_pdf[i]=pdf_values[i]-MathExp(noncentral_t_pdf[i]);
            delta_noncentral_t_cdf[i]=(1.0-cdf_values[i])-MathExp(noncentral_t_cdf[i]);
            delta_noncentral_t_quantile[i]=x_values[i]-noncentral_t_quantile[i];
           }
        }
      else
      for(int i=0;i<N;i++)
        {
         delta_noncentral_t_pdf[i]=pdf_values[i]-noncentral_t_pdf[i];
         delta_noncentral_t_cdf[i]=(1.0-cdf_values[i])-noncentral_t_cdf[i];
         delta_noncentral_t_quantile[i]=x_values[i]-noncentral_t_quantile[i];
        }
     }
//--- check calculations
   for(int i=0;i<N;i++)
     {
      PrintFormat("3 %d, PDF=%.20e, CDF=%.20e, Q=%.20e,",i,noncentral_t_pdf[i],noncentral_t_cdf[i],noncentral_t_quantile[i]);
      //PrintFormat("3 %d, dPDF=%.20e, dCDF=%.20e, dQ=%.20e,",i,delta_noncentral_t_pdf[i],delta_noncentral_t_cdf[i],delta_noncentral_t_quantile[i]);
      if(MathAbs(delta_noncentral_t_pdf[i])>calc_precision_pdf)
         return false;
      if(MathAbs(delta_noncentral_t_cdf[i])>calc_precision_cdf)
         return false;
      if(MathAbs(delta_noncentral_t_quantile[i])>calc_precision_quantile)
         return false;
     }

//--- test 4: test generators
   PrintFormat("%s: Test 4: Generators of random values",test_name);
//--- generate random values and calculate 4 moments
   int random_data_count=10000;
   double values_rnd[];
   if(MathRandomNoncentralT(nu,delta,random_data_count,values_rnd)==true)
     {
      //--- calculate 4 moments of generated values
      double mean=MathMean(values_rnd,0,random_data_count);
      double variance=MathVariance(values_rnd,0,random_data_count);
      double skewness=MathSkewness(values_rnd,0,random_data_count);
      double kurtosis=MathKurtosis(values_rnd,0,random_data_count);
      //--- inital values
      double noncentral_t_mean=0;
      double noncentral_t_variance=0;
      double noncentral_t_skewness=0;
      double noncentral_t_kurtosis=0;

      //PrintFormat("        mean=%.10f,           variance=%.10f           skewness=%.10f           kurtosis=%.10f",mean,variance,skewness,kurtosis);
      //--- calculate theoretical values and compare with generated
      if(MathMomentsNoncentralT(nu,delta,noncentral_t_mean,noncentral_t_variance,noncentral_t_skewness,noncentral_t_kurtosis,error_code))
        {
         //PrintFormat("Noncentral T mean=%.10f, Noncentral T variance=%.10f  Noncentral T skewness=%.10f  Noncentral T kurtosis=%.10f",noncentral_t_mean,noncentral_t_variance,noncentral_t_skewness,noncentral_t_kurtosis);
         //PrintFormat("delta mean=%.4f, delta variance=%.4f  delta skewness=%.4f  delta kurtosis=%.4f",mean-noncentral_t_mean,variance-noncentral_t_variance,skewness-noncentral_t_skewness,kurtosis-noncentral_t_kurtosis);
        }
      else
         return false;
     }
   else
      return false;
//--- successful
   PrintFormat("%s passed",test_name);
   return true;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ShowPrecisionInfo(const string distribution,precision_info &precision)
  {
   PrintFormat("%s: PDF max error          : %5.20e",distribution,precision.pdf_max_delta);
   PrintFormat("%s: CDF max error          : %5.20e",distribution,precision.cdf_max_delta);
   PrintFormat("%s: Quantile max error     : %5.20e",distribution,precision.quantile_max_delta);
   PrintFormat("%s: PDF correct digits     : %d",distribution,GetCorrectDigits(precision.pdf_max_delta));
   PrintFormat("%s: CDF correct digits     : %d",distribution,GetCorrectDigits(precision.cdf_max_delta));
   PrintFormat("%s: Quantile correct digits: %d",distribution,GetCorrectDigits(precision.quantile_max_delta));
  }
//+------------------------------------------------------------------+
//| TestDistributions                                                |
//+------------------------------------------------------------------+
void TestDistributions(int &tests_performed,int &tests_passed)
  {
   precision_info precision;
   string test_name="";
//--- Noncentral T Distribution
   tests_performed++;
   test_name="Noncentral T distribution test";
   if(TestNoncentralTDistribution(test_name,10.0,8.0,noncentral_t_x_values,noncentral_t_pdf_values,noncentral_t_cdf_values,precision))
     {
      tests_passed++;
      ShowPrecisionInfo(test_name,precision);
     }
   else
      PrintFormat("%s failed",test_name);
   return;
  }
//+------------------------------------------------------------------+
//| UnitTests()                                                      |
//+------------------------------------------------------------------+
void UnitTests(const string package_name)
  {
   PrintFormat("Unit tests for Package %s\n",package_name);
//--- initial values
   int tests_performed=0;
   int tests_passed=0;
//--- test distributions
   TestDistributions(tests_performed,tests_passed);
//--- print statistics
   PrintFormat("\n%d of %d passed",tests_passed,tests_performed);
  }
//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
  {
   UnitTests("Stat");
  }
//+------------------------------------------------------------------+
